from . import placeholder_sets as placeholder_set
from . import placeholders as placeholder
from .core import (
  DynamicResponseBuilder,
  DynamicResponseError,
  DynamicResponseMissingPlaceholderError,
  DynamicResponsePlaceholderInvalidReturnError,
  DynamicResponseRecursionError,
  DynamicResponseSyntaxError,
)

__all__ = (
  'DynamicResponseBuilder',
  'placeholder',
  'placeholder_set',
)



import os
import pathlib as p
import sys
from collections.abc import Callable
from sys import exit

import glfw as gf
import imgui
import OpenGL.GL as gl
from imgui.integrations.glfw import GlfwRenderer

from ..src.tcr_console import console as c


def ctrlc_grace_exit() -> bool:
  print('^C', end='')
  return True  # Exit


class ImGuiHandler:
  frame: Callable[[], None]
  """The frame callback."""
  title: str
  """The title of the window."""
  font_path: p.Path | None
  """The path to the font file to use, or None to disable."""
  ctrlc_handler: Callable[[], bool] | None
  """The Ctrl-C handler to use, or None to disable. Return True to exit the gui, False to continue. Do not exit()."""

  def __init__(
    self,
    frame: Callable[['ImGuiHandler'], None],
    *,
    title: str = 'Untitled window',
    font_path: p.Path | str | None = None,
    ctrlc_handler: Callable[[], bool] = ctrlc_grace_exit,
    start_maximised: bool = False,
  ):
    if not callable(frame):
      raise TypeError('frame must be Callable[[], None]')
    self.frame = frame
    if ctrlc_handler is not None and not callable(ctrlc_handler):
      raise TypeError('ctrlc_handler must be Callable[[], bool]')
    self.ctrlc_handler = ctrlc_handler

    if not isinstance(title, str):
      raise TypeError('title must be str')
    self.title = title

    if isinstance(font_path, str):
      font_path = p.Path(font_path)
    if font_path is not None and not isinstance(font_path, p.Path):
      raise TypeError('font_path must be str or pathlib.Path (or None to disable)')
    if font_path is not None and not font_path.is_file():
      raise FileNotFoundError(f"font_path {font_path} doesn't exist or is not a file")
    self.font_path = font_path

    self.start_maximised = start_maximised

    self._setup()

  def _setup(self):
    imgui.create_context()
    if not gf.init():
      raise RuntimeError('Could not initialize OpenGL context')

    gf.window_hint(gf.CONTEXT_VERSION_MAJOR, 3)
    gf.window_hint(gf.CONTEXT_VERSION_MINOR, 3)
    gf.window_hint(gf.OPENGL_PROFILE, gf.OPENGL_CORE_PROFILE)
    gf.window_hint(gf.OPENGL_FORWARD_COMPAT, gl.GL_TRUE)

    window = gf.create_window(*gf.get_monitor_physical_size(gf.get_primary_monitor()), self.title, None, None)
    gf.make_context_current(window)

    if not window:
      gf.terminate()
      raise RuntimeError('Could not initialize GLFW window.')

    if self.start_maximised:
      gf.maximize_window(window)

    impl = GlfwRenderer(window)

    io = imgui.get_io()
    jb = io.fonts.add_font_from_file_ttf(self.font_path, 30) if self.font_path is not None else None
    impl.refresh_font_texture()

    while not gf.window_should_close(window):
      try:
        self._render(impl, window, jb)
      except KeyboardInterrupt:
        if self.ctrlc_handler():
          break

    impl.shutdown()
    gf.terminate()

  def _render(self, impl, window, font):
    gf.poll_events()
    impl.process_inputs()
    imgui.new_frame()

    gl.glClearColor(0.1, 0.1, 0.1, 1)
    gl.glClear(gl.GL_COLOR_BUFFER_BIT)

    if font is not None:
      imgui.push_font(font)

    self.frame()

    if font is not None:
      imgui.pop_font()

    imgui.render()
    impl.render(imgui.get_draw_data())
    gf.swap_buffers(window)



def ensure_dependencies():
  MISSING_MODULES = []
  try:
    import imgui
    from imgui.integrations.glfw import GlfwRenderer
  except ModuleNotFoundError:
    MISSING_MODULES.append('imgui[glfw]')

  try:
    import glfw as gf
  except ModuleNotFoundError:
    MISSING_MODULES.append('glfw')

  try:
    import OpenGL.GL as gl
  except ModuleNotFoundError:
    MISSING_MODULES.append('PyOpenGL')

  try:
    import numpy as np
  except ModuleNotFoundError:
    MISSING_MODULES.append('numpy')

  if MISSING_MODULES:
    from ..src.tcr_console import console as c

    c.error('Please pip install the following modules for the GUI mode to work.')
    c.debug(*MISSING_MODULES, withprefix=False)
    raise RuntimeError('Missing modules.')



from typing import Any, Self


class Singleton:
  """Keep single instance at all times.

  ## Make sure to put the `tcr.Singleton` inheretance before all others!
  ### ❌ `class A(B, tcr.Singleton)`
  ### ✅ `class A(tcr.Singleton, B)`

  ```py
  class SingleTuple(tcr.Singleton, tuple):
      ...

    tup = SingleTuple(('tup1',))
    tup2 = SingleTuple(('tup2',)) # Those arguments are effectively ignored

    console(tup)  # -> ('tup1',)
    console(tup2) # -> ('tup1',)
  ```
  """

  __singleton_instance__ = None

  def __new__(cls, *args, **kwargs) -> Self:
    if cls.__singleton_instance__ is None:
      cls.__singleton_instance__ = super().__new__(cls, *args, **kwargs)
    return cls.__singleton_instance__


class NoInit:
  def __init__(self, *args, **kwargs) -> None:
    msg = 'You cannot instiantiate this class'

    if hasattr(self, '__noinit_msg'):
      msg = str(getattr(self, '__noinit_msg'))

    raise RuntimeError(msg)


def partial_class(_f=None, /, *, add_as_kwargs_to_init: bool = True) -> type:
  """### Simillarly to functools.partial allows subclasses of decorated class to hold a meta-like state to be later used in e.g. root's init or anywhere else.

  ```py
  @tcr.partial_class
  class Button:
    def __init__(self, label: str):
      self.label = label

    def __repr__(self):
      return f'Button(label={self.label!r})'

  class YesButton(Button, label='Yes'): ...
  class NoButton(Button, label='No'): ...

  y = YesButton()
  c(repr(y)) # -> Button(label='Yes')
  n = NoButton()
  c(repr(n)) # -> Button(label='No')
  n2 = NoButton(label='Nuh uh') # Overridden
  c(repr(n2)) # -> Button(label='Nuh uh')
  ```

  Args:
    add_as_kwargs_to_init: bool = True. If True, the state  will be passed as keyword arguments into `__init__`.
  """

  def decorator(cls: type) -> type:
    def new_init_subclass(new_cls, **kwargs) -> None:
      new_cls._partial_state = kwargs

      if not add_as_kwargs_to_init:
        return

      original_init = new_cls.__init__

      def new_init(self, *args, **init_kwargs):
        merged_kwargs = {**self._partial_state, **init_kwargs}  # Merge both states such that caller can override it, like functools.partial
        original_init(self, *args, **merged_kwargs)

      new_cls.__init__ = new_init

    cls.__init_subclass__ = classmethod(new_init_subclass)
    return cls

  if _f is not None:
    return decorator(_f)
  return decorator
